package org.example

data class State(
        val matched: String,
        val remaining: String,
)

data class AvailableState(val seq: Sequence<State> = emptySequence()) : Sequence<State> by seq {
  val isEmpty: Boolean
    get() = seq.none()
}

class MatchResult(val available: AvailableState) {

  val isSuccess: Boolean
    get() = !this.available.isEmpty

  val isFailure: Boolean
    get() = this.available.isEmpty

  override fun toString(): String {
    return if (isSuccess) {
      "MatchResult(success)"
    } else {
      "MatchResult(failure)"
    }
  }
}

// 재귀 하향 분석기.
interface RegexItem {
  override fun toString(): String
  fun findMatch(str: String): AvailableState
}

fun RegexItem.match(item: String): MatchResult {
  // 기본 매칭 함수. AvailableState를 MatchResult로 변환
  return MatchResult(this.findMatch(item))
}

class AndThenItem(val left: RegexItem, val right: RegexItem) : RegexItem {
  override fun toString(): String = "${left}${right}"
  override fun findMatch(str: String): AvailableState {
    val leftMatch = left.findMatch(str)
    if (leftMatch.isEmpty) {
      return AvailableState() // If left match fails, return empty sequence
    }
    // If left match is successful, try to match the right item with the remaining string
    // from the left match.

    return AvailableState(
            leftMatch.flatMap { state ->
              val rightMatch = right.findMatch(state.remaining)
              if (!rightMatch.isEmpty) {
                // If right match is successful, combine the matched parts
                rightMatch.map { rightState ->
                  State(state.matched + rightState.matched, rightState.remaining)
                }
              } else {
                // If right match fails, return an empty sequence
                emptySequence()
              }
            }
    )
  }
}

class CharItem(val value: String) : RegexItem {
  override fun toString(): String = value
  override fun findMatch(str: String): AvailableState {
    return when {
      // 첫번째 문자가 value와 일치하는지 확인
      str.isNotEmpty() && str[0].toString() == value -> {
        AvailableState(sequenceOf(State(value, str.substring(1))))
      }
      else -> AvailableState()
    }
  }
}

fun matchMany(
        str: String,
        item: RegexItem,
): Sequence<State> {
  // 욕심쟁이 매칭을 위한 헬퍼 함수
  return item.findMatch(str).seq.flatMap { state ->
    if (state.remaining.isEmpty()) {
      sequenceOf(state) // If remaining is empty, return the matched state
    } else {
      // Otherwise, continue matching with the remaining string
      matchMany(state.remaining, item).map { nextState ->
        State(state.matched + nextState.matched, nextState.remaining)
      } + sequenceOf(state) // Include the current state as well
    }
  }
}

// fun matchMany(str: String, item: RegexItem): Sequence<State> = sequence {
//   val stack = mutableListOf(item.findMatch(str).seq)
//   while (stack.isNotEmpty()) {
//     val current = stack.removeAt(stack.lastIndex)
//     for (state in current) {
//       yield(state) // Yield the current state
//       if (state.remaining.isNotEmpty()) {
//         // If there is remaining string, continue matching
//         stack.add(item.findMatch(state.remaining).seq.map { nextState ->
//           State(state.matched + nextState.matched, nextState.remaining)
//         })
//       }
//     }
//   }
// }

class PlusItem(val item: RegexItem) : RegexItem {
  override fun toString(): String = "${item}+"
  override fun findMatch(str: String): AvailableState {
    return AvailableState(matchMany(str, item))
  }
}

class StarItem(val item: RegexItem) : RegexItem {
  override fun toString(): String = "${item}*"
  override fun findMatch(str: String): AvailableState {
    // *는 0개 이상의 매칭을 의미하므로, 먼저 시도해보고 실패하면 빈 시퀀스를 반환
    val matchResult = this.item.findMatch(str)
    if (matchResult.isEmpty) {
      // If the item does not match, return an empty sequence
      return AvailableState(sequenceOf(State("", str)))
    }
    // If it matches, return the successful match and continue matching with the remaining string
    return AvailableState(
            matchResult.flatMap { state ->
              sequenceOf(state) + matchMany(state.remaining, this.item)
            }
    )
  }
}

class QuestionItem(val item: RegexItem) : RegexItem {
  override fun toString(): String = "${item}?"
  override fun findMatch(str: String): AvailableState {
    // ?는 0개 또는 1개 매칭을 의미하므로, 먼저 시도해보고 실패하면 빈 시퀀스를 반환
    val matchResult = this.item.findMatch(str)
    if (matchResult.isEmpty) {
      // If the item does not match, return an empty sequence
      return AvailableState(sequenceOf(State("", str)))
    }
    // If it matches, return the successful match
    return AvailableState(matchResult.map { State(it.matched, it.remaining) })
  }
}

class DotItem : RegexItem {
  override fun toString(): String = "."
  override fun findMatch(str: String): AvailableState =
          // .은 임의의 한 문자와 매칭되므로, 첫 문자가 존재하면 매칭 성공
          when {
            str.isNotEmpty() ->
                    AvailableState(sequenceOf(State(str[0].toString(), str.substring(1))))
            else -> AvailableState() // 빈 문자열에 대해서는 매칭 실패
          }
}

class AlternationItem(val left: RegexItem, val right: RegexItem) : RegexItem {
  override fun toString(): String = "(${left}|${right})"
  override fun findMatch(str: String): AvailableState {
    // Alternation은 왼쪽 또는 오른쪽 항목 중 하나와 매칭되므로, 각각 시도해보고 성공하는 경우를 반환
    val leftMatch = left.findMatch(str)
    val rightMatch = right.findMatch(str)

    return AvailableState(
            (leftMatch + rightMatch) // 두 매칭 결과를 합쳐서 반환
    )
  }
}
